/* $Id$ */

/*
 * Copyright (c) 2006 Grigory Batalov <bga at altlinux org>
 * Copyright (c) 2004 Darren Tucker <dtucker at zip com au>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "includes.h"

#ifdef USE_ADJTIMEX
#include <sys/timex.h>
#include <errno.h>
#ifdef adjtime
# undef adjtime
#endif

#include "ntpd.h"

/* scale factor used by adjtimex freq param.  1 ppm = 65536 */
#define ADJTIMEX_FREQ_SCALE 65536

int
_compat_adjtime(const struct timeval *delta, struct timeval *olddelta)
{
	static struct timeval tlast = {0,0};
	struct timeval tnow, tdelta;
	double adjust, adjust_last, interval, error_ppm;
	long freq_delta = 0, tick_delta = 0;
	struct timex tmx;
	int result, saved_errno;
	/* Get system clock ticks per second */
	long hz = sysconf(_SC_CLK_TCK);

	gettimeofday(&tnow, NULL);
	adjust = (double)delta->tv_sec;
	adjust += (double)delta->tv_usec / 1000000;

	/* Even if the caller doesn't care about the olddelta, we do */
	if (olddelta == NULL)
		olddelta = &tdelta;

	/* Adjust time in a simple way */
	result = adjtime(delta, olddelta);
	saved_errno = errno;

	/* How much correction from last call */
	adjust_last  = (double)olddelta->tv_sec;
	adjust_last += (double)olddelta->tv_usec / 1000000;
	
	/* Get current adjtimex values */
	tmx.modes = 0;
	if (adjtimex(&tmx) == -1)
		log_warn("adjtimex get failed");

	/* If we can calculate interval */
	if (tlast.tv_sec && tlast.tv_usec) {
		interval = (double)(tnow.tv_sec - tlast.tv_sec);
		interval += (double)(tnow.tv_usec - tlast.tv_usec) / 1000000;

		log_debug("interval %0.3lf olddelta %0.3lf (delta - olddelta) %0.3lf",
			interval, adjust_last, adjust - adjust_last);

		if (interval != 0) {
			/* Calculate error in "parts per million" */
			error_ppm = (adjust - adjust_last) * 1000000 / interval;
			/* Half of error_ppm leads to less fluctiations */
			error_ppm /= 2;
			/* Tick adjustment */
			tick_delta = error_ppm / hz;
			/* Frequency adjustment */
			freq_delta = ADJTIMEX_FREQ_SCALE * (error_ppm - tick_delta * hz);

			log_debug("error_ppm %0.3lf freq_delta %ld tick_delta %ld",
				error_ppm, freq_delta, tick_delta);

			/* Adjust the kernel skew.  */
			tmx.freq += freq_delta;
			tmx.tick += tick_delta;
			/* Check frequency bounds: -tolerance <= tolerance */
			if (tmx.freq < -tmx.tolerance) {
				tmx.tick--;
				tmx.freq += ADJTIMEX_FREQ_SCALE * hz;
			}
			else if (tmx.freq > tmx.tolerance) {
				tmx.tick++;
				tmx.freq -= ADJTIMEX_FREQ_SCALE * hz;
			}
			/* Check tick bounds: 900000/hz <= 1100000/hz */
			if (tmx.tick < 900000/hz)
				tmx.tick = 900000/hz;
			else if (tmx.tick > 1100000/hz)
				tmx.tick = 1100000/hz;

			tmx.modes = ADJ_FREQUENCY | ADJ_TICK;

			if (olddelta->tv_sec == 0 && olddelta->tv_usec == 0) {
				tmx.maxerror = 0;
				tmx.status &= ~STA_UNSYNC;
				tmx.modes |= ADJ_MAXERROR | ADJ_STATUS;
			}
			if (adjtimex(&tmx) == -1)
				log_warn("adjtimex set freq failed");
		}

	}
	else
		log_debug("skiping very first adjtimex");
	tlast = tnow;
	errno = saved_errno;
	return result;
}
#endif
